"""Working with the Product data over the REST Web Services:  CRUD operations, Objects, and Fragments

Generic Data Service introduces an approach to work with the Product data over the REST Web Services.
The Service delivers the basic methods to be able to fulfill any operation with the data like CRUD-operations (Create-Read-Update-Delete).
The methods are generic, which means, it fully relies on the state of the Product Schema and allows to work with any Configuration item or
Data Definition presented in the Product Schema. That approach allows extending the Schema with new entities, or customize the present
ones, and then immediately dealing with new artifacts via Web Services.
"""



import json
import requests
from matrix42sdk.AuthNClient import Matrix42RestClient
from requests.exceptions import HTTPError


class ObjectsDataService(Matrix42RestClient):
    """Objects (/api/data/objects) which addresses operations with the Objects (instances of the Configuration Items)

    The Service provides methods for CRUD operations (Create-Read-Update-Delete) for Configuration Items presented in the Schema.

    Args:
        Matrix42RestClient ([type]): Inherit Matrix42RestClient object

    `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/API%3A_Generic_Data_Service>`_
    """

    def __init__(self, _path=None, _full_header=None, **kwargs):
        super().__init__(**kwargs)
        self._path = "/M42Services/api/data/objects"
        self._full_header = self.get_matrix42_access_header()

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        self._path = value

    def get_object(self, ciName, objectId, full="true"):
        """Gets the whole Object with the specified Configuration Item name and object ID.

        The Service returns exclusively the data belonged to object (e.g. attributes, N:1 relations).
        Virtual attributes, like N:M or 0/1 to N relations are not returned and need to be obtained
        with the additional Web Service request.

        TimeStamp is a special marker is used for the concurrency issues handling on modifying the Fragment.
        TimeStamp is generated for each Object Fragment.

        Args:
            ciName (str):
                Required. Technical name of the Configuration Item (e.g. for Incident is "SPSActivityTypeIncident")
            objectId (str):
                Required. Id of the Object of specified Configuration Item
            full (str):
                Optional. Signals to load the whole Object with all related multi-fragments data, otherwise, all multi-fragments are omitted.
                While Rest API defaults to 'false', this SDK does 'true'.

        Returns:
            The whole Object with all defined Data Definitions and attributes. The
            Multi-fragments are present only when parameter full is set.

        `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Objects_Data_Service%3A_Get_Object_data>`_
        """
        # full=true is important for getting complete object, including version
        req_url = self.url + self.path + "/%s/%s?full=%s" % (ciName, objectId, full)
        try:

            r_ci_get = requests.get(
                req_url, verify=self._ssl_verify, headers=self._full_header
            )
            r_ci_get.raise_for_status()

            if r_ci_get.status_code == 404:
                return Exception(
                    "The object with the specified Configuration Item and Object ID is not present, or not allowed for the caller."
                )

            return json.loads(r_ci_get.text)

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")

        except Exception as err:
            print(f"Other error occurred: {err}")

    def update_object(self, ciName, jsonBody, full="true"):
        """Updates the object of the specified Configuration Item name and object ID.

        The method works in pair with Get Object method, which delivers the original
        Object with the current Timestamp (market of current state), which is used for
        concurrency tracking.

        Returns:
            The server does not return any data. Nonetheless, we return the Request's object.

        Args:
            ciName (str):
                Required. Technical name of the Configuration Item (e.g. for Incident is "SPSActivityTypeIncident")
            objectid (str):
                Required. Id of the Object of specified Configuration Item
            full (str):
                Optional. Signals to load the whole Object with all related multi-fragments data, otherwise, all multi-fragments are omitted.
                While Rest API defaults to 'false', this SDK does 'true'.

        `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Object_Data_Service%3A_Update_Object>`_

        """
        # true => full update | must be same in the get method
        put_url = self.url + self.path + "/%s?full=%s" % (ciName, full)

        try:
            r_ci_update = requests.put(
                put_url, verify=self._ssl_verify, headers=self._full_header, json=jsonBody
            )
            r_ci_update.raise_for_status()

            if r_ci_update.status_code == 500:
                return Exception(
                    "Concurrency. The object you want to update has been updated by another process!"
                )

            return r_ci_update

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")

        except Exception as err:
            print(f"Other error occurred: {err}")

    def create_object(self, ciName, jsonBody):
        """Creates a new Object of the specified Configuration Item.

        If needed it is possible explicitly set IDs for the created Object or related fragments.
        If IDs are omitted in the Request, they are auto-generated

        Returns:
            Guid. ID of the Created Object

        Args:
            ciName (str):
                Technical name of the Configuration Item (e.g. for Incident is "SPSActivityTypeIncident")

        `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Object_Data_Service%3A_Create_Object>`_
        """
        put_url = self.url + self.path + "/%s" % ciName
        try:

            r_ci_create = requests.post(
                put_url, verify=self._ssl_verify, headers=self._full_header, json=jsonBody
            )
            r_ci_create.raise_for_status()

            if r_ci_create.status_code == 401:
                return Exception("Unauthorized")
            elif r_ci_create.status_code == 415:
                return Exception("The request Content-type is not defined")
            elif r_ci_create.status_code == 500:
                return Exception(
                    "Internal Server Error. E.g. wrong reference, or mandatory attribute is missing"
                )
            else:
                return r_ci_create

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")

        except Exception as err:
            print(f"Other error occurred: {err}")

    def delete_object(self, ciName, objectId):
        """Deletes the object from Database defined by the Configuration Item name and the object ID.

        Returns:
            The server does not return any data. Nonetheless, we return the Request's object.

        Args:
            ciName (str):
                Required. Technical name of the Configuration Item (e.g. for Incident is "SPSActivityTypeIncident")
            objectId (str):
                Required. Id of the Object of specified Configuration Item

        `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Objects_Data_Service%3A_Delete_Object>`_
        """
        put_url = self.url + self.path + "/%s/%s" % (ciName, objectId)
        try:

            r_ci_delete = requests.delete(
                put_url, verify=self._ssl_verify, headers=self._full_header
            )
            r_ci_delete.raise_for_status()

            return r_ci_delete

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")

        except Exception as err:
            print(f"Other error occurred: {err}")



import os
import requests
from matrix42sdk.Exceptions import AuthNError


MATRIX42_GENERATE_ACCESS_TOKEN_ENDPOINT = (
    "/m42Services/api/ApiToken/GenerateAccessTokenFromApiToken/"
)


class Matrix42RestClient(object):
    """Main Authentication Class

    Args:
        hostname (str): Full URL of ESM server, e.g. "https://matrix.firm.com"
        api_token (str): Generated API Token in the ESM GUI. Used for getting Access Token
        ssl_verify (bool): Requests' parameter for wherther TLS is being checked. Use it if hostname is using
                            Self-Signed Certificates
    """

    def __init__(
        self,
        _url=None,
        _api_token=None,
        _ssl_verify=False,
        _configKeyType=None,
    ):

        self._headers = dict({"Content-Type": "application/json"})
        self._ssl_verify = _ssl_verify
        self._url = _url
        self._configKeyType = _configKeyType
        self.__uses_shell = True
        self.__uses_app_token = True

        MATRIX42SDK_API_TOKEN = os.environ.get("MATRIX42SDK_API_TOKEN", None)
        MATRIX42_URL = os.environ.get("MATRIX42_URL", None)

        if MATRIX42_URL is None:
            self._url = _url.rstrip("/")
        elif MATRIX42_URL is not None:
            self._url = MATRIX42_URL.rstrip("/")
        else:
            raise AuthNError("URL is missing")

        if MATRIX42SDK_API_TOKEN is None:
            self._api_token = _api_token
        elif MATRIX42SDK_API_TOKEN is not None:
            self._api_token = MATRIX42SDK_API_TOKEN
        else:
            raise AuthNError(
                "All methods require to use either username/password or API Token."
            )

    def __str__(self):
        for attribute, value in self.__dict__.items():
            print(attribute, " => ", value)
        return ""

    @property
    def ssl_verify(self):
        return self._ssl_verify

    @ssl_verify.setter
    def ssl_verify(self, value):
        self._ssl_verify = value

    @property
    def api_token(self):
        return self._api_token

    @api_token.setter
    def api_token(self, value):
        self._api_token = value

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @classmethod
    def set_access_token_header(cls, _headers, _url, _api_token, _ssl_verify, *args, **kwargs):

        site_endpoint = _url + MATRIX42_GENERATE_ACCESS_TOKEN_ENDPOINT

        # dont use .update() on dict because it is updated in place, hence returning None type
        tkn_dict = dict({"Authorization": "Bearer %s" % _api_token})
        access_headers = dict(_headers, **tkn_dict)
        postReq = requests.post(site_endpoint, headers=access_headers, verify=_ssl_verify, *args, **kwargs)
        raw_acc_token = postReq.json()["RawToken"]

        # here we want to update in-place
        acc_tkn = dict({"Authorization": "Bearer %s" % raw_acc_token})
        _headers.update(**acc_tkn)
        return cls(_headers)

    def get_matrix42_access_header(self):
        if self.__uses_app_token is True:
            if self.__uses_shell is True:
                Matrix42RestClient.set_access_token_header(
                    self._headers, self._url, self._api_token, self._ssl_verify
                )
        return self._headers



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['matrix42sdk', 'matrix42sdk.api_endpoints']

package_data = \
{'': ['*']}

install_requires = \
['mypy<1.0.0', 'pytest>=6.0.0,<7.0.0', 'requests>=2,<3']

setup_kwargs = {
    'name': 'matrix42sdk',
    'version': '3.0.1',
    'description': 'Python SDK for Matrix42 Enterprise Service Management CMBD',
    'long_description': '# Python 3 Matrix42 SDK for Enterprise Service Management (ESM) tool\n\n[![Quality gate](https://sonarcloud.io/api/project_badges/quality_gate?project=dmpe_matrix42sdk)](https://sonarcloud.io/dashboard?id=dmpe_matrix42sdk)\n\n"ERP for Software Asset Management"\n\n[![matrix42sdk package in my-feed feed in Azure Artifacts](https://feeds.dev.azure.com/johnmalc/2efa647f-e5a5-4720-835d-4fc45fde9432/_apis/public/Packaging/Feeds/0c65acb4-f8ae-4df6-9f17-9db0f7687350/Packages/313927f6-82a5-4f2b-9bb0-9b90bcf3cec2/Badge)](https://dev.azure.com/johnmalc/Matrix42SDK/_packaging?_a=package&feed=0c65acb4-f8ae-4df6-9f17-9db0f7687350&package=313927f6-82a5-4f2b-9bb0-9b90bcf3cec2&preferRelease=true)\n[![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=dmpe_matrix42sdk&metric=alert_status)](https://sonarcloud.io/dashboard?id=dmpe_matrix42sdk)\n[![Build Status](https://dev.azure.com/johnmalc/Matrix42SDK/_apis/build/status/dmpe.matrix42sdk?branchName=master)](https://dev.azure.com/johnmalc/Matrix42SDK/_build/latest?definitionId=10&branchName=master)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n![GitHub](https://img.shields.io/github/license/dmpe/matrix42sdk)\n![Pull Requests](https://img.shields.io/badge/PR-Welcome-%23FF8300.svg)\n![LGTM Grade](https://img.shields.io/lgtm/grade/python/github/dmpe/matrix42sdk)\n![Requires.io](https://img.shields.io/requires/github/dmpe/matrix42sdk)\n[![Total alerts](https://img.shields.io/lgtm/alerts/g/dmpe/matrix42sdk.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/dmpe/matrix42sdk/alerts/)\n![GitHub repo size](https://img.shields.io/github/repo-size/dmpe/matrix42sdk)\n\n## The background story\n\nAs of July 2020, Matrix42 AG, a German company is offering a [Configuration Management Database (CMBD)](https://www.matrix42.com/en/digital-workspace-management/enterprise-service-management) - basically a competing product to ServiceNOW and many others like Microsoft Intune.\n\nTheir Angular based product can be installed on-prem and used as ITIL supporting tool for the company.\n\nIt provides a REST API which this Python3+ SDK tries to cover.\nUnfortunately, Swagger support is currently not available which has led me to write this client package myself - becoming my first python3 client SDK.\n\nIt shows - also due to learning Python OOP principles.\n\n### Documentation of Rest API\n\n- <https://help.matrix42.com/030_DWP/030_INT> Intro\n- <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations> How to use the API\n- <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Matrix42_Web_Services_API> Matrix42 Approach to the API\n- <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Web_Services%3A_Authentication_types> Example of AuthN\n- <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Web_Services_tokens%3A_Generate_API_Token> Generate API Token\n\nSetting up the Sphinx auto-generate documentation:\n\n- <https://samnicholls.net/2016/06/15/how-to-sphinx-readthedocs/>\n\n# What works and what does not?\n\nWorks:\n\n- get and put (i.e. update) fragment\n\nUse Cases:\n\n- updating CI version numbers\n\nSemi/or not working:\n\n- creating fragment was not tested, but is implemented as a Rest API call\n- get and put object -> here not tested extensively and one must expect bugs\n\n# Testing\n\nWithout having a Matrix42 ESM portal which allows API access and is publicly available for unlimited use,\nthe only option how to test this library is to use **your** own ESM portal installation (be it in public cloud or on prem).\nHence the need to use your own API keys, etc.\n\nFrom this follows that this library cannot - as of now - include more python tests, at least not publicly available.\nIt is my recommendation to write your `own, private tests` and report issues here.\n\nSorry for inconvenience!\n\n# Usage\n\n## Python 3 - Simple Testing\n\nUsing `requests` you can call\n\n```\nimport requests\n\nurl = "https://xxxx/m42Services/api/ApiToken/GenerateAccessTokenFromApiToken/"\n\npayload = {}\nheaders = {\n  \'Content-Type\': \'application/json\',\n  \'Authorization\': \'Bearer the \'short\' version of bearer token generated from Administration Panel in the GUI\'\n}\n\nresponse = requests.request("POST", url, headers=headers, data = payload, verify = False)\n\nprint(response.text.encode(\'utf8\'))\n\n# only then start using proper API requests\nimport requests\n\nurl = "https://xxxx/m42Services/api/data/fragments/Ud_SoftwareproduktVersionClassBase/775c82cf-d243-4bfb-a1b2-f3edad93c826"\n\npayload = {}\nheaders = {\n  \'Authorization\': \'Bearer "output of the previous requests "RawToken" \'\n}\n\nresponse = requests.request("GET", url, headers=headers, data = payload)\n\nprint(response.text.encode(\'utf8\'))\n```\n## Matrix42 SDK for Python\n\nTo use Matrix42 SDK, first decide which authN approach you are going to use.\n\nFor a **basic**, create a `Matrix42RestClient` object by using your (or any other matrix42 CMBD) account:\n\n```{python}\nimport matrix42sdk\nfrom matrix42sdk import AuthNClient\nfrom matrix42sdk.api_endpoints import matrix42_objects, matrix42_fragments\n```\n\nFor using `Access/API Tokens`, you can set your `MATRIX42SDK_API_TOKEN` via a shell (**higher priority**):\n\n```{shell}\nexport MATRIX42_URL="xxx"\nexport MATRIX42SDK_API_TOKEN="xxx"\n```\n\nand then:\n\n```{python3}\nmat = matrix42_fragments.GDSFragements()\n```\n\nThen to [get a fragment for a specific CI](https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Fragments_Data_Service%3A_Get_Fragment_data), insert correct parameters according to the documentation:\n\n```\nJUPYTERLAB_ID_FRAG = "8c51cfff-bf16-452e-8d2c-527cc25518c3"\nSYS_ENTITY = "SPSSoftwareType"\nfull_ci_frg = mat.get_fragement(SYS_FRAGEMENT, JUPYTERLAB_ID_FRAG)\n```\n\n## Documentation\n\nWe use <https://sphinx-rtd-theme.readthedocs.io/> and <https://www.sphinx-doc.org/en/master/>\n\n```\ncd docs\nsphinx-apidoc -o source ../matrix42sdk\nmake html\n```\n\n# Building this package yourself\n\nThis package is being build on Azure DevOps services:\n\n<https://dev.azure.com/johnmalc/Matrix42SDK/_build>\n\nand further uses SonarCloud:\n\n<https://sonarcloud.io/dashboard?id=dmpe_matrix42sdk>\n\nUse to install package from non-PyPI feed (the azure one):\n\n```\npip3 install --index https://pkgs.dev.azure.com/johnmalc/Matrix42SDK/_packaging/my-feed/pypi/simple/ matrix42sdk\n```\n\n## PyPI\n\n```\npoetry build\n```\n',
    'author': 'John Malc',
    'author_email': 'cincenko@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dmpe/matrix42sdk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import json
import requests
from matrix42sdk.AuthNClient import Matrix42RestClient
from requests.exceptions import HTTPError


class FragmentsDataService(Matrix42RestClient):
    """Fragments (/api/data/fragments), provides the operation for working with the Fragments (Instances of the Data Definitions)

    The Service provides methods for CRUD operations (Create-Read-Update-Delete) for Data Definitions presented in the Schema.

    Args:
        Matrix42RestClient ([type]): Inherit Matrix42RestClient object

    `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/API%3A_Generic_Data_Service>`_
    """

    def __init__(self, _path=None, _full_header=None, **kwargs):
        super().__init__(**kwargs)
        self._path = "/M42Services/api/data/fragments"
        self._full_header = self.get_matrix42_access_header()

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        self._path = value

    def delete_fragement(self, ddname, fragmentId):
        """Deletes the fragment from Database defined by the Data Definition name and the object ID.

        The operation is required for cases of multi-fragments or optional fragments.

        Args:
            ddname (str):
                The technical name of the Data Definition (e.g. SPSActivityClassBase)

            fragmentId (str):
                Id of the Fragment of specified Data Definition

        `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Fragments_Data_Service:_Delete_Fragment>`_
        """
        req_url = self.url + self.path + "/%s/%s" % (ddname, fragmentId)
        try:
            r_ci_del = requests.delete(
                req_url, verify=self._ssl_verify, headers=self._full_header
            )
            r_ci_del.raise_for_status()
            return r_ci_del

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")

        except Exception as err:
            print(f"Other error occurred: {err}")

    def get_fragment(self, ddname, fragmentId):
        """Reads the whole Fragment of the specified Data Definition with defined Id.

        Returns:
            Returns the JSON object with all attributes specified in the Data Definition.
            The Service returns null if the fragment with the specified fragmentId does not exist,
            or the Object the Fragment belongs to, is not allowed for the caller.

        Args:
            ddname (str):
                The technical name of the Data Definition (e.g. SPSActivityClassBase)

            fragmentId (str):
                Id of the Fragment of specified Data Definition

        `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Fragments_Data_Service%3A_Get_Fragment_data>`_

        """
        # full=true is important for getting complete object, including version
        req_url = self.url + self.path + "/%s/%s?full=true" % (ddname, fragmentId)
        try:
            r_ci = requests.get(
                req_url, verify=self._ssl_verify, headers=self._full_header
            )
            r_ci.raise_for_status()
            return json.loads(r_ci.text)

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")

        except Exception as err:
            print(f"Other error occurred: {err}")

    # https://docs.python.org/3/tutorial/controlflow.html#special-parameters
    # only allows keyword arguments as indicated
    def get_fragments_list(
        self,
        ddname,
        *,
        where=None,
        columns=None,
        pageSize=None,
        pageNumber=None,
        sort=None,
        includeLocalizations=None,
    ):
        """Retrieves a list of fragments with a defined list of columns, which match the specified search criteria.

        Returns:
            Retrieves a list of fragments with a defined list of columns which match the specified
            search criteria and are sorted in the defined order. In case of need, the method returns
            the schema metadata for the returned data.

            An array of JSON objects where each object represents the Data Definition fragment which matches
            the specified search criteria and is allowed for the current user (caller).

        Args:
            ddname (str):
                Required. The technical name of the Data Definition (e.g. SPSActivityClassBase)
            where (str):
                Optional. A-SQL Where Expression. If no filtering expression is specified, the Service returns all fragments
                of the specified Data Definition which are allowed for the caller.
            columns (str):
                Optional. A-SQL Column expression defines the columns in the result set, separated by Comma. If no Columns defines,
                then the Operation returns only Fragment Ids. Example: Name, Parent.Name as ParentName
            pageSize (int):
                Optional. Sets the number of records (fragments) returned by Operation. Used in combination with the PageNumber parameter
            pageNumber (int):
                Optional. Sets the Number of the page. Used in combination with `pageSize` property.
            sort (str):
                Optional. Defines the sorting in the result. Example: "Name ASC, CreatedDate DESC"
            includeLocalizations (bool):
                Optional. Specifies if the Localization table should be included in the result. The parameter works ONLY
                with the request with "schema-info" directive. The service handles correctly only attributes of the
                requested Data Definition. Related attributes which are part of the response, but not
                in Data Definition (e.g. SPSActivityClassBase: Category.Name) keep only values of the request culture.
                If you need all localizations for these attributes as well, please run a dedicated request to Data
                Definition which keeps this attribute (e.g. SPSScCategoryClassBase)

        `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Fragments_Data_Service%3A_Get_a_list_of_Fragments>`_

        """
        req_url = self.url + self.path + "/%s" % ddname

        payload = dict()
        if where is not None:
            payload.update({"where": where})
        if columns is not None:
            payload.update({"columns": columns})
        if pageSize is not None:
            payload.update({"pageSize": pageSize})
        if pageNumber is not None:
            payload.update({"pageNumber": pageNumber})
        if sort is not None:
            payload.update({"sort": sort})
        if includeLocalizations is not None:
            payload.update({"includeLocalizations": includeLocalizations})

        try:
            r_ci_list = requests.get(
                req_url,
                params=payload,
                verify=self._ssl_verify,
                headers=self._full_header,
            )
            r_ci_list.raise_for_status()
            return json.loads(r_ci_list.text)

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")

        except Exception as err:
            print(f"Other error occurred: {err}")

    def get_fragment_relations_list(
        self,
        ddname,
        fragmentId,
        relationName,
        *,
        where=None,
        columns=None,
        pageSize=None,
        pageNumber=None,
        sort=None,
        includeLocalizations=None,
    ):
        """Retrieves a list of fragment's relations with a defined list of columns which match the specified search
        criteria and are sorted in the defined order.

        Returns:
            Retrieves a list of fragments with a defined list of columns which match the specified
            search criteria and are sorted in the defined order. In case of need, the method returns
            the schema metadata for the returned data.

            An array of JSON objects where each object represents the Data Definition fragment which matches
            the specified search criteria and is allowed for the current user (caller).

        Args:
            ddname (str):
                Required. The technical name of the Data Definition (e.g. SPSActivityClassBase)
            fragmentId (str):
                Required. Id of the Fragment of specified Data Definition
            relationName (str):
                Required. Technical name of the relation (e.g. AttachedUsers).
            where (str):
                Optional. A-SQL Where Expression. If no filtering expression is specified, the Service returns all fragments
                of the specified Data Definition which are allowed for the caller.
            columns (str):
                Optional. A-SQL Column expression defines the columns in the result set, separated by Comma. If no Columns defines,
                then the Operation returns only Fragment Ids. Example: Name, Parent.Name as ParentName
            pageSize (int):
                Optional. Sets the number of records (fragments) returned by Operation. Used in combination with the PageNumber parameter
            pageNumber (int):
                Optional. Sets the Number of the page. Used in combination with `pageSize` property.
            sort (str):
                Optional. Defines the sorting in the result. Example: "Name ASC, CreatedDate DESC"
            includeLocalizations (bool):
                Optional. Specifies if the Localization table should be included in the result. The parameter works ONLY
                with the request with "schema-info" directive. The service handles correctly only attributes of the
                requested Data Definition. Related attributes which are part of the response, but not
                in Data Definition (e.g. SPSActivityClassBase: Category.Name) keep only values of the request culture.
                If you need all localizations for these attributes as well, please run a dedicated request to Data
                Definition which keeps this attribute (e.g. SPSScCategoryClassBase)

        `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Fragments_Data_Service%3A_Get_a_list_of_Fragment_Relations>`_

        """
        req_url = self.url + self.path + "/%s/%s/%s" % (ddname, fragmentId, relationName)

        payload = dict()
        if where is not None:
            payload.update({"where": where})
        if columns is not None:
            payload.update({"columns": columns})
        if pageSize is not None:
            payload.update({"pageSize": pageSize})
        if pageNumber is not None:
            payload.update({"pageNumber": pageNumber})
        if sort is not None:
            payload.update({"sort": sort})
        if includeLocalizations is not None:
            payload.update({"includeLocalizations": includeLocalizations})

        try:
            r_ci_list = requests.get(
                req_url,
                params=payload,
                verify=self._ssl_verify,
                headers=self._full_header,
            )
            r_ci_list.raise_for_status()
            return json.loads(r_ci_list.text)

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")

        except Exception as err:
            print(f"Other error occurred: {err}")

    def create_fragment(self, ddname, fragmentData):
        """Creates a new Data Definition fragment. The operation is required for cases of multi-fragments or optional fragments.

        Args:
            ddname (str):
                Required. The technical name of the Data Definition (e.g. SPSActivityClassBase)
            fragmentData: (json)
                Request Body with a JSON containing all necessary data for Fragment creation.
                For JSON  structure examples see  Fragments Data Service: Get Fragment page.

        Returns:
            Id of the created Fragment.

        `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Fragments_Data_Service%3A_Create_Fragment>`_

        Untested
        """
        put_url = self.url + self.path + "/%s" % ddname
        try:
            r_ci_create = requests.post(
                put_url, verify=self._ssl_verify, headers=self._full_header, data=fragmentData
            )
            r_ci_create.raise_for_status()
            return r_ci_create

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")

        except Exception as err:
            print(f"Other error occurred: {err}")

    def update_fragment(self, ddname, jsonBody):
        """Updates the specified Data Definition fragment attributes.

        The Service modifies only attributes which are explicitly specified in the Request Body.
        The attributes which are not mentioned in the request are not affected by the Update operation.

        Needs jsonBody which has "JSON Object with fragment attributes with new values".

        Use :meth:`get_fragment <matrix42sdk.FragmentsDataService.get_fragment>` JSON object to update fields.

        JSON object should contain:
            ID (str): Attribute with the updated Fragment Id value is required.
            Concurrency tracking (str): if you want to track concurrency also include the TimeStamp attribute.
            Reset the attribute value (str): in case you want to reset the attribute value add this attribute to JSON object with the value null.

        Returns:
            The system returns no data, but this method returns Requests' full `put` object.

        `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Fragments_Data_Service%3A_Update_fragment>`_

        """
        # true => full update | must be same in the get method
        put_url = self.url + self.path + "/%s?full=true" % ddname

        try:
            r_ci_update = requests.put(
                put_url, verify=self._ssl_verify, headers=self._full_header, data=jsonBody
            )
            r_ci_update.raise_for_status()

            if r_ci_update.status_code == 204:
                print("Update of CI fragment has been ok")

            return r_ci_update

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")

        except Exception as err:
            print(f"Other error occurred: {err}")

    def delete_fragment_relation(
        self, ddname, fragmentId, relationName, relationFragmentId
    ):
        """Deletes the relation from Database defined by Data Definition name, fragment ID and Relation name.

        The operation is required for managing many-to-many relations.

        Returns:
            The system returns no data, but this method returns Requests' full `post` object.

        Args:
            ddname (str):
                The technical name of the Data Definition (e.g. SPSActivityClassBase)
            fragmentid (str):
                Id of the Fragment of specified Data Definition
            relationName (str):
                Technical name of the relation (e.g. AttachedUsers).
            relationFragmentId (str):
                Id of the Data Definition's fragment that is added as a relation.

        `Matrix42 URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Fragments_Data_Service%3A_Delete_Fragment_Relation>`_

        """
        req_url = (
            self.url
            + self.path
            + "/%s/%s/%s/%s" % (ddname, fragmentId, relationName, relationFragmentId)
        )
        try:
            r_ci_delete_rel = requests.delete(
                req_url, verify=self._ssl_verify, headers=self._full_header
            )
            r_ci_delete_rel.raise_for_status()
            return r_ci_delete_rel

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")

        except Exception as err:
            print(f"Other error occurred: {err}")

    def add_fragment_relation(self, ddname, fragmentId, relationName, relationFragmentId):
        """Adds relation to the Database defined by Data Definition name, fragment ID and Relation name.

        The operation is required for managing many-to-many relations.
        The operation adds a single object to relation. If you need to add multiple relations to the object,
        you need to make a call of the Service for each of them

        Returns:
            The system returns no data, but this method returns Requests' full `post` object.

        Args:
            ddname (str):
                The technical name of the Data Definition (e.g. SPSActivityClassBase)
            fragmentid (str):
                Id of the Fragment of specified Data Definition
            relationName (str):
                Technical name of the relation (e.g. AttachedUsers).
            relationFragmentId (str):
                Id of the Data Definition's fragment that is added as a relation.

        `URL <https://help.matrix42.com/030_DWP/030_INT/Business_Processes_and_API_Integrations/Public_API_reference_documentation/Fragments_Data_Service%3A_Add_Fragment_Relation>`_

        """
        req_url = (
            self.url
            + self.path
            + "/%s/%s/%s/%s" % (ddname, fragmentId, relationName, relationFragmentId)
        )
        try:
            r_ci_add_rel = requests.get(
                req_url, verify=self._ssl_verify, headers=self._full_header
            )
            r_ci_add_rel.raise_for_status()
            return r_ci_add_rel

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")

        except Exception as err:
            print(f"Other error occurred: {err}")
